import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

public class Algorithms_TSP {
	static double current_x, current_y;
	static String[] statuses; //"unserved", "served"
	
	static double solveInstance(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_TSP instance, int lookahead){
		double length = 0;
		current_x = 0;
		current_y = 0; 
		int T = instance.T;
		statuses = new String[T];
		for(int t = 0; t < T; t++)
			statuses[t] = "unserved";
		for(int t = 0; t < T; t++)
			length += visitLocation(algorithm, gradualLossFactor, maxGradualHorizon, instance, lookahead, t);				
		return length;
	}
	
	static double visitLocation(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_TSP instance, int lookahead, int t_current){
		int indexOfNextLocation = -1;
		if(algorithm.equals("NN"))
			indexOfNextLocation = getIndexOfLocation_NN(instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIP"))
			indexOfNextLocation =  getIndexOfLocation_SnapshotMIP(false, gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIPForecasts"))
			indexOfNextLocation = getMostFrequentIndexOfLocation(gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		double stepLength = distanceBetween(current_x, current_y, instance.x[indexOfNextLocation], instance.y[indexOfNextLocation]);
		current_x = instance.x[indexOfNextLocation];
		current_y = instance.y[indexOfNextLocation];
		statuses[indexOfNextLocation] = "served";
		return stepLength;
	}
	
	static int getMostFrequentIndexOfLocation(double gradualLossFactor, int maxGradualHorizon, Instance_TSP instance, int lookahead, int t_current){
		int numberOfShootings = 25;
		int[] indizesOfLocation = new int[numberOfShootings];
		for(int it = 0; it < numberOfShootings; it++)
			indizesOfLocation[it] = getIndexOfLocation_SnapshotMIP(true, gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		return (new Distribution(indizesOfLocation)).getValueWithHighestProbability();
	}

	static int getIndexOfLocation_NN(Instance_TSP instance, int lookahead, int t_current){
		int index = -1;
		double nearestDistance = 1000;
		for(int t = 0; t < Math.min(t_current + lookahead, instance.T); t++)
			if(statuses[t].equals("unserved") && distanceBetween(current_x, current_y, instance.x[t], instance.y[t]) < nearestDistance){
				index = t;
				nearestDistance = distanceBetween(current_x, current_y, instance.x[t], instance.y[t]);
			}
		return index;
	}
	
	static int getIndexOfLocation_SnapshotMIP(boolean forecasts, double gradualLossFactor, int maxGradualHorizon, Instance_TSP instance, int lookahead, int t_current){
		int n_lookahead = Math.min(lookahead, instance.T - t_current);
		int n_forecasts;
		if(!forecasts)
			n_forecasts = 0;
		else
			n_forecasts = Math.min(instance.T - t_current - n_lookahead, maxGradualHorizon);
		int n = n_lookahead + n_forecasts;
		int[] indizes = new int[n];
		int i_current = 0;
		for(int t = 0; t < Math.min(t_current + lookahead, instance.T); t++)
			if(statuses[t].equals("unserved") && i_current < n){
				indizes[i_current] = t;
				i_current++;
			}
		if(forecasts){
			for(int t = Math.min(t_current + lookahead, instance.T); t < instance.T; t++)
				if(i_current < n){
					indizes[i_current] = t;
					i_current++;
				}
			instance.createForecasts(gradualLossFactor, t_current, lookahead);
		}
		double[][] c = new double[n][n];
		for(int i = 0; i < n; i++)
			for(int j = 0; j < n; j++)
				if(i < n_lookahead && j < n_lookahead)
					c[i][j] = distanceBetween(instance.x[indizes[i]], instance.y[indizes[i]], instance.x[indizes[j]], instance.y[indizes[j]]);
				else if(i >= n_lookahead && j >= n_lookahead)
					c[i][j] = distanceBetween(instance.x_forecasted[i], instance.y_forecasted[i], instance.x_forecasted[j], instance.y_forecasted[j]);
				else if(i < n_lookahead && j >= n_lookahead)
					c[i][j] = distanceBetween(instance.x[indizes[i]], instance.y[indizes[i]], instance.x_forecasted[j], instance.y_forecasted[j]);
				else if(i >= n_lookahead && j < n_lookahead)
					c[i][j] = distanceBetween(instance.x_forecasted[i], instance.y_forecasted[i], instance.x[indizes[j]], instance.y[indizes[j]]);
		double[] c_start = new double[n];
		for(int j = 0; j < n; j++){
			if(j < n_lookahead)
				c_start[j] = distanceBetween(current_x, current_y, instance.x[indizes[j]], instance.y[indizes[j]]);
			else
				c_start[j] = 999999;
		}		

		int indexOfLocation = -1;
		try{
			IloCplex cplex = new IloCplex();
			IloNumVar[][] x = new IloNumVar[n][];
			for(int i = 0; i < n; i++)
				x[i] = cplex.boolVarArray(n);
			IloNumVar[] x_start = cplex.boolVarArray(n);
			IloNumVar[] x_end = cplex.boolVarArray(n);
			IloNumVar[] T_mip = cplex.numVarArray(n, 0, 999);
			
			IloLinearNumExpr objective = cplex.linearNumExpr();
			for(int i = 0; i < n; i++)
				for(int j = 0; j < n; j++)
					objective.addTerm(c[i][j], x[i][j]);
			for(int j = 0; j < n; j++)
				objective.addTerm(c_start[j], x_start[j]);
	        cplex.addMinimize(objective);
	        
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int j = 0; j < n; j++)
	        		if(j != i)
	        			expr.addTerm(1, x[i][j]);
	        	expr.addTerm(1, x_end[i]);
	        	cplex.addEq(expr, 1);
	        }
	        
	        for(int j = 0; j < n; j++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int i = 0; i < n; i++)
	        		if(i != j)
	        			expr.addTerm(1, x[i][j]);
	        	expr.addTerm(1, x_start[j]);
	        	cplex.addEq(expr, 1);
	        }
	        
	        IloLinearNumExpr expr1 = cplex.linearNumExpr();
	        for(int j = 0; j < n; j++)
	        	expr1.addTerm(1, x_start[j]);
	        cplex.addEq(expr1, 1);
	        
	        IloLinearNumExpr expr2 = cplex.linearNumExpr();
	        for(int i = 0; i < n; i++)
	        	expr2.addTerm(1, x_end[i]);
	        cplex.addEq(expr2, 1);
	        
	        for(int i = 0; i < n; i++)
	        	for(int j = 0; j < n; j++){
	        		IloLinearNumExpr expr = cplex.linearNumExpr();
        			expr.addTerm(1, T_mip[i]);
        			expr.addTerm(-1, T_mip[j]);
        			expr.addTerm(999, x[i][j]);
        			cplex.addLe(expr, 998);
	        	}        	
	        cplex.setOut(null);
	        cplex.setParam(IloCplex.DoubleParam.EpGap, 0.05);
	        cplex.setParam(IloCplex.DoubleParam.TiLim, 60);
	        cplex.solve();
	        for(int j = 0; j < n_lookahead; j++)
	        	if(Math.abs(cplex.getValue(x_start[j]) - 1) < 0.0001)
	        		indexOfLocation = indizes[j];
	        cplex.end();
		}
	    catch (IloException ex){
	    	System.out.println("Concert Error: " + ex);
	    }
		return indexOfLocation;
	}
	
	static double distanceBetween(double x1, double y1, double x2, double y2){
		return Math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
	}
}